from typing import NewType

import z3
from pysmt.shortcuts import GE, LE, Real, And, GT, LT, Equals, Or

from DPGM import Node
from distribution import Distribution
from solver_instance import SolverInstance
from z3_instance import Z3Instance

# label constants
Label = NewType("Label", str)
l_in = Label("in")
l_out = Label("out")
l_undec = Label("undec")


class LabelingScheme:

    num_args = 0  # The number of arguments each labeling scheme takes

    def get_label(self, argument: Node, distribution: Distribution) -> Label:
        """
        Get the label of the given argument according the given model and the labeling approach.
        :return: l_in, l_out or l_undec
        """
        # implemented by subclasses
        pass

    def get_constraints(self, node: Node, si: SolverInstance):
        """
        Return a list of constraints on the probabilistic variable of the node where each constraint corresponds to a
        different label. I.e. for cautious labeling and node A, return [p_A == 1, p_A == 0, And(p_A < 1, p_A > 0)].
        Used to compute all labelings.
        """
        # implemented by subclasses
        return []

    def get_constraints_z3(self, node: Node, z3i: Z3Instance):
        """
        Return a list of constraints on the probabilistic variable of the node where each constraint corresponds to a
        different label. I.e. for cautious labeling and node A, return [p_A == 1, p_A == 0, And(p_A < 1, p_A > 0)].
        Used to compute all labelings.
        """
        # implemented by subclasses
        return []


class Threshold(LabelingScheme):
    """
    Given two thresholds t_in and t_out, all nodes with probability
     - greater equal t_in are labeled 'in',
     - less than or equal to t_out are labeled 'out',
     - and between t_in and t_out are labeled 'undec'.
     """
    num_args = 2

    def __init__(self, t_in: float, t_out: float):
        self.t_in = t_in
        self.t_out = t_out

    def get_label(self, node: Node, distribution: Distribution):
        prob = distribution.get_marginal_probability(node)
        if prob >= self.t_in:
            return l_in
        elif prob <= self.t_out:
            return l_out
        else:
            return l_undec

    def get_constraints(self, node: Node, si: SolverInstance):
        node_prob = si.get_node_symbol(node)
        in_constraint = GE(node_prob, Real(self.t_in))
        out_constraint = LE(node_prob, Real(self.t_out))
        undec_constraint = And(LT(node_prob, Real(self.t_in)), GT(node_prob, Real(self.t_out)))
        return [in_constraint, out_constraint, undec_constraint]

    def get_constraints_z3(self, node: Node, z3i: Z3Instance):
        node_prob = z3i.get_node_prob_var(node)
        in_constraint = node_prob >= self.t_in
        out_constraint = node_prob <= self.t_out
        undec_constraint = z3.And(node_prob < self.t_in, node_prob > self.t_out, z3i.context)
        return [in_constraint, out_constraint, undec_constraint]


class Cautious(Threshold):
    """
    Arguments with probability 1 are labeled 'in', those with probability 0 are labeled 'out' and all others are
    labeled 'undec'.
    """
    num_args = 0

    def __init__(self):
        super().__init__(1.0, 0.0)


class Firm(LabelingScheme):
    """
    Arguments with probability 0.5 are labeled 'undec', those with probability >0.5 are labeled 'in' and all those with
    probability <0.5 are labeled 'out'.
    """
    def get_label(self, node: Node, distribution: Distribution):
        prob = distribution.get_marginal_probability(node)
        if prob > 0.5:
            return l_in
        elif prob < 0.5:
            return l_out
        else:
            return l_undec

    def get_constraints(self, node: Node, si: SolverInstance):
        node_prob = si.get_node_symbol(node)
        in_constraint = GT(node_prob, Real(0.5))
        out_constraint = LT(node_prob, Real(0.5))
        undec_constraint = Equals(node_prob, Real(0.5))
        return [undec_constraint, in_constraint, out_constraint]
        # undec first because checking for equality is probably easier than inequalities

    def get_constraints_z3(self, node: Node, z3i: Z3Instance):
        node_prob = z3i.get_node_prob_var(node)
        in_constraint = node_prob > 0.5
        out_constraint = node_prob < 0.5
        undec_constraint = node_prob == 0.5
        return [undec_constraint, in_constraint, out_constraint]
        # undec first because checking for equality is probably easier than inequalities


class ThresholdClassicalOld(LabelingScheme):
    """
    Requires two thresholds t_in and t_min_attack_out. A node is labeled
     - 'in' if its probability is greater or equal t_in,
     - 'out' if it is not labeled 'in' and one of its attackers has probability greater or equal t_min_attack_out, and
     - 'undec' in all other cases.

     Uses old implementation of get_constraints_z3 which does not necessarily produce unique labelings, though might be
     faster in some cases.
     """
    num_args = 2

    def __init__(self, t_in: float, t_min_attack_out: float):
        self.t_in = t_in
        self.t_att_out = t_min_attack_out

    def get_label(self, node: Node, distribution: Distribution):
        node_prob = distribution.get_marginal_probability(node)
        if node_prob >= self.t_in:
            return l_in
        else:
            label = l_undec
            for attacker in node.get_parents():
                attacker_prob = distribution.get_marginal_probability(attacker)
                if attacker_prob >= self.t_att_out:
                    label = l_out
                    break
            return label

    def get_constraints(self, node: Node, si: SolverInstance):
        node_prob = si.get_node_symbol(node)
        if self.t_in == self.t_att_out:
            in_constraint = GE(node_prob, Real(self.t_in))
            not_in_constraint = LT(node_prob, Real(self.t_in))
            return [in_constraint, not_in_constraint]
        else:
            t_max = max(self.t_in, self.t_att_out)
            t_min = min(self.t_in, self.t_att_out)
            first_constraint = GE(node_prob, Real(t_max))
            second_constraint = LT(node_prob, Real(t_min))
            third_constraint = And(LT(node_prob, Real(t_max)), GE(node_prob, Real(t_min)))
            return [first_constraint, second_constraint, third_constraint]

    def get_constraints_z3(self, node: Node, z3i: Z3Instance):
        node_prob = z3i.get_node_prob_var(node)
        if self.t_in == self.t_att_out:
            in_constraint = node_prob >= self.t_in
            not_in_constraint = node_prob < self.t_in
            return [in_constraint, not_in_constraint]
        else:
            t_max = max(self.t_in, self.t_att_out)
            t_min = min(self.t_in, self.t_att_out)
            first_constraint = node_prob >= t_max
            second_constraint = node_prob < t_min
            third_constraint = z3.And(node_prob < t_max, node_prob >= t_min, z3i.context)
            return [first_constraint, second_constraint, third_constraint]


class ClassicalOld(ThresholdClassicalOld):
    """
    A node is labeled
     - 'in' if it holds with probability 1,
     - 'out' if it is not labeled 'in' and one of its attackers holds with probability 1, and
     - 'undec' in all other cases.

     Uses old implementation of get_constraints_z3 which does not necessarily produce unique labelings, though might be
     faster in some cases.
    """
    num_args = 0

    def __init__(self):
        super().__init__(1.0, 1.0)


class ThresholdClassical(LabelingScheme):
    """
    Requires two thresholds t_in and t_min_attack_out. A node is labeled
     - 'in' if its probability is greater or equal t_in,
     - 'out' if it is not labeled 'in' and one of its attackers has probability greater or equal t_min_attack_out, and
     - 'undec' in all other cases.
     """
    num_args = 2

    def __init__(self, t_in: float, t_min_attack_out: float):
        self.t_in = t_in
        self.t_att_out = t_min_attack_out

    def get_label(self, node: Node, distribution: Distribution):
        node_prob = distribution.get_marginal_probability(node)
        if node_prob >= self.t_in:
            return l_in
        else:
            label = l_undec
            for attacker in node.get_parents():
                attacker_prob = distribution.get_marginal_probability(attacker)
                if attacker_prob >= self.t_att_out:
                    label = l_out
                    break
            return label

    def get_constraints(self, node: Node, si: SolverInstance):
        node_prob = si.get_node_symbol(node)
        in_constraint = GE(node_prob, Real(self.t_in))
        parent_out_constraint = Or(
            [GE(si.get_node_symbol(parent), Real(self.t_att_out)) for parent in node.get_parents()])
        out_constraint = And(GT(node_prob, Real(self.t_in)), parent_out_constraint)
        parent_undec_constraint = And(
            [LT(si.get_node_symbol(parent), Real(self.t_att_out)) for parent in node.get_parents()])
        undec_constraint = And(LT(node_prob, Real(self.t_in)), parent_undec_constraint)
        return [in_constraint, out_constraint, undec_constraint]

    def get_constraints_z3(self, node: Node, z3i: Z3Instance):
        node_prob = z3i.get_node_prob_var(node)
        in_constraint = node_prob >= self.t_in
        parent_out_constraint = z3.Or(
            [z3i.get_node_prob_var(parent) >= self.t_att_out for parent in node.get_parents()])
        out_constraint = z3.And(node_prob < self.t_in, parent_out_constraint)
        parent_undec_constraint = z3.And(
            [z3i.get_node_prob_var(parent) < self.t_att_out for parent in node.get_parents()])
        undec_constraint = z3.And(node_prob < self.t_in, parent_undec_constraint)

        return [in_constraint, out_constraint, undec_constraint]


class Classical(ThresholdClassical):
    """
    A node is labeled
     - 'in' if it holds with probability 1,
     - 'out' if it is not labeled 'in' and one of its attackers holds with probability 1, and
     - 'undec' in all other cases.
    """
    num_args = 0

    def __init__(self):
        super().__init__(1.0, 1.0)


class Optimistic(LabelingScheme):
    """
    A node is labeled
     - 'in' if it holds with probability larger than 0, and
     - 'out' otherwise.
     """

    def get_label(self, node: Node, distribution: Distribution):
        prob = distribution.get_marginal_probability(node)
        if prob > 0.0:
            return l_in
        else:
            return l_out

    def get_constraints(self, node: Node, si: SolverInstance):
        node_prob = si.get_node_symbol(node)
        in_constraint = GT(node_prob, Real(0))
        out_constraint = Equals(node_prob, Real(0))
        return [in_constraint, out_constraint]

    def get_constraints_z3(self, node: Node, z3i: Z3Instance):
        node_prob = z3i.get_node_prob_var(node)
        in_constraint = node_prob > 0
        out_constraint = node_prob == 0
        return [in_constraint, out_constraint]


class Pessimistic(LabelingScheme):
    """
    A node is labeled
     - 'in' if it holds with probability 1, and
     - 'out' otherwise.
     """

    def get_label(self, node: Node, distribution: Distribution):
        prob = distribution.get_marginal_probability(node)
        if prob == 1.0:
            return l_in
        else:
            return l_out

    def get_constraints(self, node: Node, si: SolverInstance):
        node_prob = si.get_node_symbol(node)
        in_constraint = Equals(node_prob, Real(1))
        out_constraint = LT(node_prob, Real(1))
        return [in_constraint, out_constraint]

    def get_constraints_z3(self, node: Node, z3i: Z3Instance):
        node_prob = z3i.get_node_prob_var(node)
        in_constraint = node_prob == 1
        out_constraint = node_prob < 1
        return [in_constraint, out_constraint]


###################
# Helper functions
###################

def get_labeling_scheme_class(name: str):
    """
    By dark magic, get the labeling scheme class given its name as string, e.g. "firm" to yield the class 'Firm'.

    :param name: the name of the labeling scheme class (case insensitive)
    :return: the labeling scheme class corresponding to the name, or None if no match was found
    """
    name = name.lower()
    queue = LabelingScheme.__subclasses__().copy()
    while queue:
        labeling_class = queue.pop()
        if name == labeling_class.__name__.lower():
            return labeling_class
        queue.extend(labeling_class.__subclasses__())
    return None


def get_all_labeling_scheme_names():
    """
    :return: a list of the names of all available labeling schemes
    """
    names = []
    queue = LabelingScheme.__subclasses__().copy()
    while queue:
        labeling_class = queue.pop()
        queue.extend(labeling_class.__subclasses__())
        names.append(labeling_class.__name__)
    return names



import os

import DPGM


def parse_interval(string, obj):
    """Parse a value or an interval and add it to either the node or edge passed as obj.
    '0.2:0.8' -> add interval (0.2, 0.8)
    '0.5' -> add value 0.5
    'ABC' -> print warning, add nothing"""

    split = string.split(':')

    p1 = float(split[0])
    if len(split) == 1:
        obj.add_value(p1)
    elif len(split) == 2:
        p2 = float(split[1])
        obj.add_interval((p1, p2))
    else:
        print("Warning: Ill-formed probability '" + string + "'")


def parse_tgf(path: str):
    """
    Parse a file in (extended) trivial graph format. Returns a directed probabilistic graphical model (DPGM) object.
    """
    with open(path, "r") as file:
        tgf_file = file.read()

    base = os.path.basename(path)  # 'AFs\simple_odd_cycle.tgf' -> 'simple_odd_cycle.tgf'
    (name, extension) = os.path.splitext(base)  # name: 'simple_odd_cycle', extension: '.tgf'
    dpgm = DPGM.DPGM(name, path)

    # remove blank lines and line comments (starting with ';')
    lines = tgf_file.split('\n')
    node_declarations = []
    edge_declarations = dict()  # from edge_type to list of edge declarations
    start_edges = False
    current_edge_type = None
    for line in lines:
        split = line.split(";")
        if split[0].startswith("#"):
            start_edges = True
            current_edge_type = split[0].strip('#').strip()
            edge_declarations[current_edge_type] = []
            continue
        if split[0].strip() != "":
            if start_edges:
                edge_declarations[current_edge_type].append(split[0])
            else:
                node_declarations.append(split[0])

    # parse nodes and edges
    for d in node_declarations:
        # format: <node_id> [<node_name>] [<node_prob>|<node_prob_min>:<node_prob_max>]
        tokens = [token for token in d.split(' ') if token.strip()]
        node_id = tokens[0]
        node = DPGM.Node(node_id)

        if len(tokens) > 1:
            next_token = tokens[1]
            if next_token[0].isalpha():  # <node_name> is given
                node.name = next_token
                if len(tokens) > 2:
                    next_token = tokens[2]
                else:
                    next_token = None
            if next_token:
                parse_interval(next_token, node)
        dpgm.add_node(node)

    for edge_type in edge_declarations.keys():
        for d in edge_declarations[edge_type]:
            tokens = [token for token in d.split(' ') if token.strip()]
            node_from_id = tokens[0]
            node_from = dpgm.get_node(node_from_id)
            node_to_id = tokens[1]
            node_to = dpgm.get_node(node_to_id)
            if edge_type is not "":
                edge = DPGM.Edge(node_from, node_to, edge_type=edge_type)
            else:
                edge = DPGM.Edge(node_from, node_to)
            if len(tokens) > 2:
                next_token = tokens[2]
                if next_token[0].isalpha():
                    edge.label = next_token
                    if len(tokens) > 3:
                        next_token = tokens[3]
                    else:
                        next_token = None
                if next_token:
                    parse_interval(next_token, edge)
            dpgm.add_edge(edge)

    return dpgm



from typing import Dict, List, Tuple
from fractions import Fraction

import z3

from assignments import Assignment
import assignments as asg
from DPGM import DPGM, Node, Edge
from distribution import Distribution


class Z3Instance:
    """Class to bundle the declared probabilistic variables, the underlying full joint distribution."""

    def __init__(self, af: DPGM, assume_independence=False):
        self.af = af
        self.assume_independence = assume_independence  # when this flag is True, no full joint probability is built

        self.full_joint_prob_vars: Dict[str, z3.ArithRef] = dict()
        self.full_joint_assignments: List[Assignment] = []
        self.other_prob_vars: Dict[str, z3.ArithRef] = dict()
        self.real_vars: Dict[str, z3.ArithRef] = dict()
        self.constraints = set()

        self.edge_prob_vars: Dict[Edge, z3.ArithRef] = dict()
        self.edge_constraints = set()
        self.edge_prob_vars_generated = False

        # create a z3 solver
        self.context = z3.Context()
        self.solver = z3.Solver(ctx=self.context)

        if not self.assume_independence:
            self._generate_full_prob_vars()
            # # serialize the constraints for the full joint distribution
            # path = os.path.splitext(self.af.path)[0]  # strip the .tgf extension
            # path1 = path + "_full-joint-probvars.py"
            # with open(path1, "w") as f:
            #     f.write(repr(self.full_joint_prob_vars))
            # print("Written to", path1)
            # path2 = path + "_constraints.py"
            # with open(path2, "w") as f:
            #     f.write(repr(self.constraints))

        # workaround to have the prob vars corresponding to the nodes' marginal probabilities in the models
        # returned by z3
        for node in self.af.get_nodes():
            self.get_node_prob_var(node)

    def generate_edge_vars(self):
        """
        Generates a probabilistic variable for each edge. If an edge has a value or interval given, those are added as
        constraints.
        """
        if self.edge_prob_vars_generated:
            # There might be several calls to generate_edge_vars, but we want to add the constraints only once.
            return
        for edge in self.af.get_edges():
            self.edge_prob_vars[edge] = z3.Real(edge.label, self.context)
            edge_prob_var = self.edge_prob_vars[edge]
            if edge.value is not None:
                self.edge_constraints.add(edge_prob_var == edge.value)
                assert edge.interval is None, "both edge.value and edge.interval where given"
            elif edge.interval is not None:
                (val_min, val_max) = edge.interval
                self.edge_constraints.add(z3.And(edge_prob_var >= val_min, edge_prob_var <= val_max, self.context))
            else:
                self.edge_constraints.add(edge_prob_var >= 0)
                self.edge_constraints.add(edge_prob_var <= 1)
        self.edge_prob_vars_generated = True

    def get_edge_var(self, edge: Edge):
        if not self.edge_prob_vars_generated:
            self.generate_edge_vars()
        return self.edge_prob_vars[edge]

    def _generate_full_prob_vars(self):
        """Adds variables for the full joint distribution to the context. Adds the constraint that they sum to one."""
        constraint = 0
        if not self.full_joint_assignments:
            self.full_joint_assignments = asg.generate(sorted(self.af.get_nodes()))
        for assignment in self.full_joint_assignments:
            assignment_prob_name = prob_name(assignment)

            prob_var = z3.Real(assignment_prob_name, self.context)
            self.full_joint_prob_vars[assignment_prob_name] = prob_var
            self.constraints.add(prob_var >= 0)
            self.constraints.add(prob_var <= 1)
            constraint += prob_var

        constraint = 1 == constraint
        self.constraints.add(constraint)

    def get_node_prob_var(self, node: Node):
        return self.get_prob_var(tuple([(node, True)]))

    def get_prob_var(self, assignment: Assignment):
        """
        Get the z3 variable corresponding to a full or partial assignment.
        If it does not exist yet, it is created along with the constraints 'new_var >= 0', 'new_var <= 1' and the
        representation of new_var as sum of the full joint variables.
        """
        assert assignment, "empty assignment given"

        prob_var_name = prob_name(assignment)
        if prob_var_name in self.other_prob_vars:
            return self.other_prob_vars[prob_var_name]
        if prob_var_name in self.full_joint_prob_vars:
            return self.full_joint_prob_vars[prob_var_name]

        prob_var = z3.Real(prob_var_name, self.context)
        self.other_prob_vars[prob_var_name] = prob_var
        self.constraints.add(prob_var >= 0)
        self.constraints.add(prob_var <= 1)
        if not self.assume_independence:
            self.constraints.add(self.sum_out(assignment, prob_var))
        return prob_var

    def get_node_value(self, node: Node, model: z3.ModelRef):
        """
        Get the value assigned to a node's probabilistic variable by the given model.
        :return: The value as float
        """
        return get_prob_var_value(self.get_node_prob_var(node), model)

    def create_real_var(self, name: str):
        """
        Creates and returns an auxiliary z3 real variable.
        """
        if name in self.real_vars:
            raise ValueError("A real var with name '" + name + "' already exists.")
        real_var = z3.Real(name, self.context)
        self.real_vars[name] = real_var
        return real_var

    def sum_out(self, part_assignment: Assignment, part_assignment_prob_var):
        """[(A,True),(B,False)], [A,B,C] -> p_A_nB = p_A_nB_C + p_A_nB_nC"""
        full_assignments = asg.sum_out(part_assignment, self.af.get_nodes())
        sum_list = []
        for assignment in full_assignments:
            assignment_name = prob_name(assignment)
            prob_var = self.full_joint_prob_vars[assignment_name]
            sum_list.append(prob_var)
        constraint = part_assignment_prob_var == z3.Sum(sum_list)
        return constraint
    
    def add_constraints(self, constraints):
        for cons in constraints:
            self.constraints.add(cons)

    def check_satisfiability(self, prob_constraints) -> z3.CheckSatResult:
        """
        Add internal and given constraints to the solver and check for a satisfying model.

        :param prob_constraints: a list of constraints
        :return: A z3 CheckSatResult, i.e. sat, unsat or unknown
        """
        self.solver.reset()
        self.solver.add(self.constraints)
        # print("self.constraints:", self.constraints)
        self.solver.add(self.edge_constraints)
        # print("self.edge_constraints", self.edge_constraints)
        self.solver.add(prob_constraints)
        # print("prob_constraints", prob_constraints)

        result = self.solver.check()
        # print(self.solver.to_smt2())
        # print(result)
        # if result == z3.sat:
        #     model = self.solver.model()
        return result

    # def print_distribution(self, model: z3.ModelRef, nodes=None, only_positive=False):
    #     """
    #     Print distribution given by the model over the assignments of the given nodes.
    #     If no list of nodes is given, the distributions over all nodes in the AF is used.
    #     """
    #     if not nodes:
    #         nodes = self.af.get_nodes()
    #     node_assignments = asg.generate(nodes)
    #     for assignment in node_assignments:
    #         prob_var = self.get_prob_var(assignment)
    #         value = get_prob_var_value(prob_var, model)
    #         if not only_positive or value > 0:
    #             asg.print_assignment_prob(assignment, value)
    #             # print(prob_var, "=", value)

    def generate_independence_constraints(self, a: List[Node], b: List[Node], c: List[Node]):
        """
        Generate constraints stating that the nodes in A are independent of the nodes in B given the nodes in C.
        For each assignment a_A, a_B and a_C, we have
                 P(a_A | a_BC) = P(a_A | a_C)
        <=> P(a_ABC) / P(a_BC) = P(a_A | a_C)
        <=>           P(a_ABC) = P(a_BC) * P(a_A | a_C)   (this would suffice if P(a_A | a_C) is known from a CPT)
        <=>           P(a_ABC) = P(a_BC) * P(a_AC) / P(a_C)
        <=>  P(a_C) * P(a_ABC) = P(a_BC) * P(a_AC)
        """
        all_nodes = sorted(a + b + c)
        constraints = []
        for assignment in asg.generate(all_nodes):
            a_ABC = assignment
            p_ABC = self.get_prob_var(a_ABC)
            a_BC: List[Tuple[Node, bool]] = [(node, v) for (node, v) in assignment if node in b or node in c]
            p_BC = self.get_prob_var(tuple(a_BC))
            a_C: List[Tuple[Node, bool]] = [(node, v) for (node, v) in assignment if node in c]
            if a_C:
                p_C = self.get_prob_var(tuple(a_C))
                a_AC: List[Tuple[Node, bool]] = [(node, v) for (node, v) in assignment if node in a or node in c]
                p_AC = self.get_prob_var(tuple(a_AC))
                constraints.append(p_C * p_ABC == p_BC * p_AC)
            else:
                # a_C is empty, i.e., the constraint is an unconditional independence constraint
                a_A: List[Tuple[Node, bool]] = [(node, v) for (node, v) in assignment if node in a]
                p_A = self.get_prob_var(tuple(a_A))
                constraints.append(p_ABC == p_BC * p_A)
        return constraints

    def distribution_from_model(self, model: z3.ModelRef) -> Distribution:
        """
        Get a distribution object from a z3 model.
        """
        distribution_dict = dict()
        for assignment in self.full_joint_assignments:
            prob_var = self.get_prob_var(assignment)
            distribution_dict[tuple(assignment)] = get_prob_var_value(prob_var, model)
        distribution = Distribution(distribution_dict, self.af.get_nodes())
        distribution.z3_model = model
        return distribution


def prob_name(assignment: Assignment) -> str:
    """
    Generates a fixed name for any assignment.
    Expects a list of tuples of nodes matched to True or False, i.e. [(node_B, False), (node_C, True)].
    Generates a name for the probability of the conjunction of the nodes, i.e. 'p_nB_C'.
    The order is determined by node IDs.
    """
    # assert list(assignment) == sorted(list(assignment))  # enforce sortedness as invariant?
    assignment = sorted(list(assignment))
    name = "p"
    for (arg, b) in assignment:
        name += "_"
        if not b:
            name += "n"
        name += arg.name.upper()  # make the name uppercase to avoid conflicts with the "n" used as prefix
    return name


def get_prob_var_value(prob_var, model: z3.ModelRef):
    """
    Get the value assigned to a probabilistic variable by the given model.
    :return: The value as float
    """
    # Based on https://stackoverflow.com/a/12600208/6620204
    value = model.eval(prob_var)
    if z3.is_int_value(value):
        return value.as_long()
    elif z3.is_rational_value(value):
        return float(value.numerator_as_long())/float(value.denominator_as_long())
    elif z3.is_algebraic_value(value):
        approx_value = value.approx(20)  # an approximation with precision 1/10^20
        return float(approx_value.numerator_as_long())/float(approx_value.denominator_as_long())
    else:
        raise ValueError("Unable to convert to float: " + str(value))


def get_prob_var_value_as_fraction(prob_var, model: z3.ModelRef) -> Fraction:
    """
    Get the value assigned to a probabilistic variable by the given model.
    :return: The value as a Fraction object
    """
    # Based on https://stackoverflow.com/a/12600208/6620204
    value = model.eval(prob_var)
    if z3.is_int_value(value):
        return Fraction(value.as_long())
    elif z3.is_rational_value(value):
        return Fraction(value.numerator_as_long(), value.denominator_as_long())
    elif z3.is_algebraic_value(value):
        approx_value = value.approx(20)
        return Fraction(approx_value.numerator_as_long(), approx_value.denominator_as_long())
    else:
        raise ValueError("Unable to convert to float: " + str(value))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cpraa']

package_data = \
{'': ['*'],
 'cpraa': ['.git/*',
           '.git/hooks/*',
           '.git/info/*',
           '.git/logs/*',
           '.git/logs/refs/heads/*',
           '.git/logs/refs/remotes/origin/*',
           '.git/objects/03/*',
           '.git/objects/04/*',
           '.git/objects/0d/*',
           '.git/objects/12/*',
           '.git/objects/14/*',
           '.git/objects/17/*',
           '.git/objects/1a/*',
           '.git/objects/1b/*',
           '.git/objects/21/*',
           '.git/objects/2d/*',
           '.git/objects/2e/*',
           '.git/objects/35/*',
           '.git/objects/39/*',
           '.git/objects/3c/*',
           '.git/objects/48/*',
           '.git/objects/4a/*',
           '.git/objects/4d/*',
           '.git/objects/50/*',
           '.git/objects/52/*',
           '.git/objects/67/*',
           '.git/objects/6a/*',
           '.git/objects/6b/*',
           '.git/objects/76/*',
           '.git/objects/7b/*',
           '.git/objects/7c/*',
           '.git/objects/7f/*',
           '.git/objects/83/*',
           '.git/objects/8d/*',
           '.git/objects/91/*',
           '.git/objects/93/*',
           '.git/objects/95/*',
           '.git/objects/9e/*',
           '.git/objects/a5/*',
           '.git/objects/a8/*',
           '.git/objects/a9/*',
           '.git/objects/af/*',
           '.git/objects/b0/*',
           '.git/objects/b2/*',
           '.git/objects/b8/*',
           '.git/objects/bd/*',
           '.git/objects/c8/*',
           '.git/objects/c9/*',
           '.git/objects/cc/*',
           '.git/objects/cd/*',
           '.git/objects/d1/*',
           '.git/objects/d3/*',
           '.git/objects/d5/*',
           '.git/objects/d9/*',
           '.git/objects/e8/*',
           '.git/objects/eb/*',
           '.git/objects/f2/*',
           '.git/objects/f3/*',
           '.git/objects/f5/*',
           '.git/objects/fc/*',
           '.git/objects/fe/*',
           '.git/objects/pack/*',
           '.git/refs/heads/*',
           '.git/refs/remotes/origin/*',
           'AFs/*',
           'AFs/KR21/*']}

install_requires = \
['PySMT>=0.9.0,<0.10.0',
 'cvxopt>=1.2.5,<2.0.0',
 'matplotlib>=3.3.4,<4.0.0',
 'numpy>=1.20.0,<2.0.0',
 'pypoman>=0.5.4,<0.6.0',
 'scipy>=1.6.0,<2.0.0',
 'z3-solver>=4.8.8,<5.0.0']

setup_kwargs = {
    'name': 'cpraa',
    'version': '0.7.0',
    'description': 'A checker for probabilistic abstract argumentation',
    'long_description': None,
    'author': 'Nikolai Käfer',
    'author_email': 'nikolai.kaefer@tu-dresden.de',
    'maintainer': 'Nikolai Käfer',
    'maintainer_email': 'nikolai.kaefer@tu-dresden.de',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
